import { type SessionStorage, type SessionData } from "@remix-run/server-runtime";
import { AuthenticateOptions, Strategy, StrategyVerifyCallback } from "remix-auth";
import type { PublicKeyCredentialDescriptorJSON } from "@simplewebauthn/types";
interface WebAuthnAuthenticator {
    credentialID: string;
    transports: string[];
}
export interface Authenticator {
    credentialID: string;
    userId: string;
    credentialPublicKey: string;
    counter: number;
    credentialDeviceType: string;
    credentialBackedUp: boolean;
    transports: string;
}
export interface UserDetails {
    id: string;
    username: string;
    displayName?: string;
}
export interface WebAuthnOptionsResponse {
    usernameAvailable: boolean | null;
    rp: {
        name: string;
        id: string;
    };
    user: {
        id: string;
        username: string;
        displayName: string;
    } | null;
    challenge: string;
    authenticators: PublicKeyCredentialDescriptorJSON[];
}
/**
 * This interface declares what configuration the strategy needs from the
 * developer to correctly work.
 */
export interface WebAuthnOptions<User> {
    /**
     * Relaying Party name - The human-readable name of your app
     */
    rpName: string | ((request: Request) => Promise<string> | string);
    /**
     * Relaying Party ID -The hostname of the website, determines where passkeys can be used
     * @link https://www.w3.org/TR/webauthn-2/#relying-party-identifier
     */
    rpID: string | ((request: Request) => Promise<string> | string);
    /**
     * Website URL (or array of URLs) where the registration can occur
     */
    origin: string | string[] | ((request: Request) => Promise<string | string[]> | string | string[]);
    /**
     * Session key to store the challenge in
     */
    challengeSessionKey?: string;
    /**
     * Return a list of authenticators associated with the user.
     * @param user object
     * @returns Authenticator
     */
    getUserAuthenticators: (user: User | null) => Promise<WebAuthnAuthenticator[]> | WebAuthnAuthenticator[];
    /**
     * Transform the user object into the shape expected by the strategy.
     * You can use a regular username, the users email address, or something else.
     * @param user object
     * @returns UserDetails
     */
    getUserDetails: (user: User | null) => Promise<UserDetails | null> | UserDetails | null;
    /**
     * Find a user in the database with their username/email.
     * @param username
     * @returns User object
     */
    getUserByUsername: (username: string) => Promise<User | null> | User | null;
    /**
     * Find an authenticator in the database by its credential ID
     * @param id
     * @returns Authenticator
     */
    getAuthenticatorById: (id: string) => Promise<Authenticator | null> | Authenticator | null;
}
/**
 * This interface declares what the developer will receive from the strategy
 * to verify the user identity in their system.
 */
export type WebAuthnVerifyParams = {
    authenticator: Omit<Authenticator, "userId">;
    type: "registration" | "authentication";
    username: string | null;
};
export declare class WebAuthnStrategy<User> extends Strategy<User, WebAuthnVerifyParams> {
    name: string;
    rpName: string | ((request: Request) => Promise<string> | string);
    rpID: string | ((request: Request) => Promise<string> | string);
    origin: string | string[] | ((request: Request) => Promise<string | string[]> | string | string[]);
    getUserAuthenticators: (user: User | null) => Promise<WebAuthnAuthenticator[]> | WebAuthnAuthenticator[];
    getUserDetails: (user: User | null) => Promise<UserDetails | null> | UserDetails | null;
    getUserByUsername: (username: string) => Promise<User | null> | User | null;
    getAuthenticatorById: (id: string) => Promise<Authenticator | null> | Authenticator | null;
    constructor(options: WebAuthnOptions<User>, verify: StrategyVerifyCallback<User, WebAuthnVerifyParams>);
    getRP(request: Request): Promise<{
        name: string;
        id: string;
        origin: string | string[];
    }>;
    generateOptions<ExtraData>(request: Request, user: User | null, extraData?: ExtraData): Promise<WebAuthnOptionsResponse & {
        extra: ExtraData extends undefined ? undefined : ExtraData;
    }>;
    private getChallenge;
    authenticate(request: Request, sessionStorage: SessionStorage<SessionData, SessionData>, options: AuthenticateOptions): Promise<User>;
}
export {};
