import {
  Strategy
} from "remix-auth";
import {
  verifyRegistrationResponse,
  verifyAuthenticationResponse
} from "@simplewebauthn/server";
import { isoBase64URL } from "@simplewebauthn/server/helpers";
class WebAuthnStrategy extends Strategy {
  constructor(options, verify) {
    super(verify);
    this.name = "webauthn";
    this.rpName = options.rpName;
    this.rpID = options.rpID;
    this.origin = options.origin;
    this.getUserAuthenticators = options.getUserAuthenticators;
    this.getUserDetails = options.getUserDetails;
    this.getUserByUsername = options.getUserByUsername;
    this.getAuthenticatorById = options.getAuthenticatorById;
  }
  async getRP(request) {
    const rp = {
      name: typeof this.rpName === "function" ? await this.rpName(request) : this.rpName,
      id: typeof this.rpID === "function" ? await this.rpID(request) : this.rpID,
      origin: typeof this.origin === "function" ? await this.origin(request) : this.origin
    };
    return rp;
  }
  async generateOptions(request, user, extraData) {
    let authenticators = [];
    let userDetails = null;
    let usernameAvailable = null;
    const rp = await this.getRP(request);
    if (!user) {
      const username = new URL(request.url).searchParams.get("username");
      if (username) {
        usernameAvailable = true;
        user = await this.getUserByUsername(username || "");
      }
    }
    if (user) {
      authenticators = await this.getUserAuthenticators(user);
      userDetails = await this.getUserDetails(user);
      usernameAvailable = false;
    }
    const crypto = await import("tiny-webcrypto");
    const options = {
      usernameAvailable,
      rp,
      user: userDetails ? { displayName: userDetails.username, ...userDetails } : null,
      challenge: isoBase64URL.fromBuffer(
        crypto.default.getRandomValues(new Uint8Array(32))
      ),
      authenticators: authenticators.map(({ credentialID, transports }) => ({
        id: credentialID,
        type: "public-key",
        transports
      })),
      extra: extraData
    };
    return options;
  }
  getChallenge(session, options) {
    var _a, _b;
    if (typeof ((_a = options.context) == null ? void 0 : _a.challenge) === "string" && ((_b = options.context) == null ? void 0 : _b.challenge) !== "") {
      return options.context.challenge;
    }
    return session.get("challenge");
  }
  async authenticate(request, sessionStorage, options) {
    var _a;
    let session = await sessionStorage.getSession(
      request.headers.get("Cookie")
    );
    try {
      let user = (_a = session.get(options.sessionKey)) != null ? _a : null;
      const rp = await this.getRP(request);
      if (request.method !== "POST")
        throw new Error("The WebAuthn strategy only supports POST requests.");
      const expectedChallenge = this.getChallenge(session, options);
      if (!expectedChallenge) {
        throw new Error(
          "Expected challenge not found. It either needs to set to the `challenge` property on the auth session, or passed as context to the authenticate function."
        );
      }
      const formData = await request.formData();
      let data;
      try {
        const responseData = formData.get("response");
        if (typeof responseData !== "string")
          throw new Error("Error");
        data = JSON.parse(responseData);
      } catch {
        throw new Error("Invalid passkey response JSON.");
      }
      const type = formData.get("type");
      let username = formData.get("username");
      if (typeof username !== "string")
        username = null;
      if (type === "registration") {
        if (!username)
          throw new Error("Username is a required form value.");
        const verification = await verifyRegistrationResponse({
          response: data,
          expectedChallenge,
          expectedOrigin: rp.origin,
          expectedRPID: rp.id
        });
        if (verification.verified && verification.registrationInfo) {
          const {
            credentialPublicKey,
            credentialID,
            counter,
            credentialBackedUp,
            credentialDeviceType
          } = verification.registrationInfo;
          const newAuthenticator = {
            credentialID,
            credentialPublicKey: isoBase64URL.fromBuffer(credentialPublicKey),
            counter,
            credentialBackedUp,
            credentialDeviceType,
            transports: ""
          };
          user = await this.verify({
            authenticator: newAuthenticator,
            type: "registration",
            username
          });
        } else {
          throw new Error("Passkey verification failed.");
        }
      } else if (type === "authentication") {
        const authenticationData = data;
        const authenticator = await this.getAuthenticatorById(
          authenticationData.id
        );
        if (!authenticator)
          throw new Error("Passkey not found.");
        const verification = await verifyAuthenticationResponse({
          response: authenticationData,
          expectedChallenge,
          expectedOrigin: rp.origin,
          expectedRPID: rp.id,
          authenticator: {
            ...authenticator,
            credentialPublicKey: isoBase64URL.toBuffer(
              authenticator.credentialPublicKey
            ),
            credentialID: authenticator.credentialID,
            transports: authenticator.transports.split(
              ","
            )
          }
        });
        if (!verification.verified)
          throw new Error("Passkey verification failed.");
        user = await this.verify({
          authenticator,
          type: "authentication",
          username
        });
      } else {
        throw new Error("Invalid verification type.");
      }
      return this.success(user, request, sessionStorage, options);
    } catch (error) {
      if (error instanceof Error) {
        return await this.failure(
          error.message,
          request,
          sessionStorage,
          options,
          error
        );
      }
      if (typeof error === "string") {
        return await this.failure(
          error,
          request,
          sessionStorage,
          options,
          new Error(error)
        );
      }
      return await this.failure(
        "Unknown error",
        request,
        sessionStorage,
        options,
        new Error(JSON.stringify(error, null, 2))
      );
    }
  }
}
export {
  WebAuthnStrategy
};
