import {
  json
} from "@remix-run/server-runtime";
import {
  Strategy
} from "remix-auth";
import {
  verifyRegistrationResponse,
  verifyAuthenticationResponse
} from "@simplewebauthn/server";
function uint8ArrayToBase64Url(uint8Array) {
  const base64String = btoa(String.fromCharCode(...uint8Array));
  return base64String.replace(/\+/g, "-").replace(/\//g, "_").replace(/=/g, "");
}
function base64UrlToUint8Array(string) {
  const base64 = string.replace(/-/g, "+").replace(/_/g, "/");
  const binaryString = atob(base64);
  const bytes = new Uint8Array(binaryString.length);
  for (let i = 0; i < binaryString.length; i++) {
    bytes[i] = binaryString.charCodeAt(i);
  }
  return bytes;
}
class WebAuthnStrategy extends Strategy {
  constructor(options, verify) {
    super(verify);
    this.name = "webauthn";
    this.rpName = options.rpName;
    this.rpID = options.rpID;
    this.origin = options.origin;
    this.getUserAuthenticators = options.getUserAuthenticators;
    this.getUserDetails = options.getUserDetails;
    this.getUserByUsername = options.getUserByUsername;
    this.getAuthenticatorById = options.getAuthenticatorById;
  }
  async getRP(request) {
    const rp = {
      name: typeof this.rpName === "function" ? await this.rpName(request) : this.rpName,
      id: typeof this.rpID === "function" ? await this.rpID(request) : this.rpID,
      origin: typeof this.origin === "function" ? await this.origin(request) : this.origin
    };
    return rp;
  }
  async generateOptions(request, sessionStorage, user) {
    let session = await sessionStorage.getSession(
      request.headers.get("Cookie")
    );
    let authenticators = [];
    let userDetails = null;
    let usernameAvailable = null;
    const rp = await this.getRP(request);
    if (!user) {
      const username = new URL(request.url).searchParams.get("username");
      if (username) {
        usernameAvailable = true;
        user = await this.getUserByUsername(username || "");
      }
    }
    if (user) {
      authenticators = await this.getUserAuthenticators(user);
      userDetails = await this.getUserDetails(user);
      usernameAvailable = false;
    }
    const crypto = await import("tiny-webcrypto");
    const options = {
      usernameAvailable,
      rp,
      user: userDetails ? { displayName: userDetails.username, ...userDetails } : null,
      challenge: uint8ArrayToBase64Url(
        crypto.default.getRandomValues(new Uint8Array(32))
      ),
      authenticators: authenticators.map(({ credentialID, transports }) => ({
        id: credentialID,
        type: "public-key",
        transports
      }))
    };
    session.set("challenge", options.challenge);
    return json(options, {
      status: 200,
      headers: {
        "Set-Cookie": await sessionStorage.commitSession(session),
        "Cache-Control": "no-store"
      }
    });
  }
  async authenticate(request, sessionStorage, options) {
    var _a;
    let session = await sessionStorage.getSession(
      request.headers.get("Cookie")
    );
    try {
      let user = (_a = session.get(options.sessionKey)) != null ? _a : null;
      const rp = await this.getRP(request);
      if (request.method !== "POST")
        throw new Error("The WebAuthn strategy only supports POST requests.");
      const expectedChallenge = session.get("challenge");
      if (!expectedChallenge)
        throw new Error(
          "Expected challenge not found. Please pass it as an option to the authenticate function."
        );
      const formData = await request.formData();
      let data;
      try {
        const responseData = formData.get("response");
        if (typeof responseData !== "string")
          throw new Error("Error");
        data = JSON.parse(responseData);
      } catch {
        throw new Error("Invalid passkey response JSON.");
      }
      const type = formData.get("type");
      let username = formData.get("username");
      if (typeof username !== "string")
        username = null;
      if (type === "registration") {
        if (!username)
          throw new Error("Username is a required form value.");
        const verification = await verifyRegistrationResponse({
          response: data,
          expectedChallenge,
          expectedOrigin: rp.origin,
          expectedRPID: rp.id
        });
        if (verification.verified && verification.registrationInfo) {
          const {
            credentialPublicKey,
            credentialID,
            counter,
            credentialBackedUp,
            credentialDeviceType
          } = verification.registrationInfo;
          const newAuthenticator = {
            credentialID: uint8ArrayToBase64Url(credentialID),
            credentialPublicKey: uint8ArrayToBase64Url(credentialPublicKey),
            counter,
            credentialBackedUp: credentialBackedUp ? 1 : 0,
            credentialDeviceType,
            transports: ""
          };
          user = await this.verify({
            authenticator: newAuthenticator,
            type: "registration",
            username
          });
        } else {
          throw new Error("Passkey verification failed.");
        }
      } else if (type === "authentication") {
        const authenticationData = data;
        const authenticator = await this.getAuthenticatorById(
          authenticationData.id
        );
        if (!authenticator)
          throw new Error("Passkey not found.");
        const verification = await verifyAuthenticationResponse({
          response: authenticationData,
          expectedChallenge,
          expectedOrigin: rp.origin,
          expectedRPID: rp.id,
          authenticator: {
            ...authenticator,
            credentialPublicKey: base64UrlToUint8Array(
              authenticator.credentialPublicKey
            ),
            credentialID: base64UrlToUint8Array(authenticator.credentialID),
            transports: authenticator.transports.split(
              ","
            )
          }
        });
        if (!verification.verified)
          throw new Error("Passkey verification failed.");
        user = await this.verify({
          authenticator,
          type: "authentication",
          username
        });
      } else {
        throw new Error("Invalid verification type.");
      }
      return this.success(user, request, sessionStorage, options);
    } catch (error) {
      if (error instanceof Error) {
        return await this.failure(
          error.message,
          request,
          sessionStorage,
          options,
          error
        );
      }
      if (typeof error === "string") {
        return await this.failure(
          error,
          request,
          sessionStorage,
          options,
          new Error(error)
        );
      }
      return await this.failure(
        "Unknown error",
        request,
        sessionStorage,
        options,
        new Error(JSON.stringify(error, null, 2))
      );
    }
  }
}
export {
  WebAuthnStrategy
};
